"""add_bootstrap_script_to_keypairs

Revision ID: ed666f476f39
Revises: d643752544de
Create Date: 2020-03-15 17:40:46.754121

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ed666f476f39'
down_revision = 'd643752544de'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('keypairs', sa.Column('bootstrap_script',
                                        sa.String(length=64 * 1024),
                                        nullable=False,
                                        server_default=''))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('keypairs', 'bootstrap_script')
    # ### end Alembic commands ###
